#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# data.py
# Copyright (c) 2020 Alvin(Xinyao) Sun <xinyao1@ualberta.ca>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import os
def readShortComplex(fileName, width=1):
    """Read binary data which in shortComplex '>i2' format

    Args:
        fileName (str): Path of the file
        width (int): Width of the data

    Example::

        >>> rslc = np.readShortComplex('sample.rslc', width=1)

    """
    return np.fromfile(fileName, '>i2').astype(np.float).view(np.complex).reshape(-1, width)


def readFloatComplex(fileName, width=1):
    """Read binary data which in floatCompoelx '>c8' format

    Args:
        fileName (str): Path of the file
        width (int): Width of the data
    """
    return np.fromfile(fileName, '>c8').astype(np.complex).reshape(-1, width)


def readFloat(fileName, width=1):
    """readFloat.

    Args:
        fileName (str): Path of the file
        width (int): Width of the data
    """
    return np.fromfile(fileName, '>f4').astype(np.float).reshape(-1, width)


def readBin(fileName, width, dataType, crop=None):
    """readBin.

    Args:
        fileName (str): Path of the file
        width (int): Width of the data
        dataType (string): 'floatComplex' | 'shortComplex' | 'float'
        crop List[int]: Crop information [crop_row, crop_col, crop_height, crop_width] 

    Returns:
        A numpy array of the data
    
    Example::

        Read RSLC in shortcomplex
        
        >>> full_rslc = readBin('example.rslc', width = 1000, dataType='shortComplex')
        >>> amp = np.abs(full_rslc)

        Read coherence in float
        >>> full_coh = readBin('example.filt.coh', width = 1000, dataType='float')

        Read a crop of interferogram in floatComplex
        >>> crop_filt = readBin('example.filt', width = 1000, dataType='floatComplex', crop=[100,200,300,300])
        >>> crop_amp = np.abs(crop_filt)
        >>> crop_phase = np.angle(crop_filt)

    """
    data_format = '>f4'
    numpy_data_format = np.float
    unit_size = 4
    mid_numpy_data_format = np.float

    if (dataType == 'floatComplex'):
        data_format = '>c8'
        numpy_data_format = np.complex
        unit_size = 8
    elif (dataType == 'shortComplex'):
        data_format = '>i2'
        numpy_data_format = np.complex
        unit_size = 4
    elif (dataType == 'float'):
        data_format = '>f4'
        numpy_data_format = np.float
        unit_size = 4

    if (unit_size > 4):
        mid_numpy_data_format = np.complex

    size_of_file = os.path.getsize(fileName)
    height = int(size_of_file / unit_size / width)

    if (crop):
        [crop_row, crop_col, crop_height, crop_width] = crop
    else:
        crop_row, crop_col, crop_height, crop_width = 0, 0, height, width

    ret = np.zeros([crop_height, crop_width], dtype=numpy_data_format)

    with open(fileName, "rb") as fin:
        for row_idx in range(crop_height):
            fin.seek(unit_size * (width * (crop_row + row_idx)+ crop_col))
            data = np.frombuffer(fin.read(unit_size * crop_width), dtype=data_format).astype(mid_numpy_data_format)
            if (dataType == 'shortComplex'):
                data = data.view(numpy_data_format)
            ret[row_idx] = data
    return ret

def writeBin(fileName, data, dataType, crop=None):
    """writeBin.

    Args:
        fileName (str): Path of the output file
        data (numpy array): Data in numpy
        dataType (string): 'floatComplex' | 'shortComplex' | 'float'

    Returns:
    
    Example::

        Write RSLC in shortcomplex
        
        >>> full_rslc = readBin('example.rslc', width = 1000, dataType='shortComplex')
        >>> writeBin('example.out.bin', dataType='shortComplex')

        Read coherence in float
        >>> full_coh = readBin('example.filt.coh', width = 1000, dataType='float')
        >>> writeBin('example.filt.coh.out.bin', dataType='float')

        Read a crop of interferogram in floatComplex
        >>> crop_filt = readBin('example.filt', width = 1000, dataType='floatComplex', crop=[100,200,300,300])
        >>> writeBin('example.filt.out.bin', dataType='floatComplex')

    """
    if (dataType == 'floatComplex'):
        writeFloatComplex(fileName, data)
    elif (dataType == 'shortComplex'):
        writeShortComplex(fileName, data)
    elif (dataType == 'float'):
        writeFloat(fileName, data)
    else:
        raise 'dataType {}, not supported'.format(dataType)

def writeShortComplex(fileName, data):
    """Write numpy data into file '>i2' format

    Data will be flatten before writing to the file

    Args:
        fileName (str): Path of the file
        data (numpy_array): Numpy array
    """
    out_file = open(fileName, 'wb')
    data.copy().view(np.float).astype('>i2').tofile(out_file)
    out_file.close()


def writeFloatComplex(fileName, data):
    """writeFloatComplex.

    Args:
        fileName (str): Path of the file
        data (numpy_array): Numpy array
    """
    out_file = open(fileName, 'wb')
    data.astype('>c8').tofile(out_file)
    out_file.close()

def writeFloat(fileName, data):
    """writeFloat.

    Args:
        fileName (str): Path of the file
        data (numpy_array): Numpy array
    """
    out_file = open(fileName, 'wb')
    data.astype('>f4').tofile(out_file)
    out_file.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File  : utils.py
# Author: Alvin(Xinyao) Sun <xinyao1@ualberta.ca>
# Date  : 25.02.2021

import numpy as np

def wrap(phase):
    """wrap.
    Return wrapped phase value
    Args:
        phase:
    """
    return np.angle(np.exp(1j*phase))




import torch
import torch.nn as nn

class DnCNN(nn.Module):
    def __init__(self, in_channels, out_channels, num_of_layers=17):
        super(DnCNN, self).__init__()
        kernel_size = 3
        padding = 1
        features = 64
        layers = []
        layers.append(nn.Conv2d(in_channels=in_channels, out_channels=features, kernel_size=kernel_size, padding=padding, bias=False))
        layers.append(nn.ReLU(inplace=True))
        for _ in range(num_of_layers-2):
            layers.append(nn.Conv2d(in_channels=features, out_channels=features, kernel_size=kernel_size, padding=padding, bias=False))
            layers.append(nn.BatchNorm2d(features))
            layers.append(nn.ReLU(inplace=True))
        layers.append(nn.Conv2d(in_channels=features, out_channels=out_channels, kernel_size=kernel_size, padding=padding, bias=False))
        self.dncnn = nn.Sequential(*layers)
    def forward(self, x):
        out = self.dncnn(x)
        return out



import torch
import torch.nn as nn
import torch.nn.functional as F

class DoubleConv(nn.Module):
    """(convolution => [BN] => ReLU) * 2"""

    def __init__(self, in_channels, out_channels, mid_channels=None, padding_mode='zeros'):
        super(DoubleConv, self).__init__()
        if not mid_channels:
            mid_channels = out_channels
        self.double_conv = nn.Sequential(
            nn.Conv2d(in_channels, mid_channels, kernel_size=3, padding=1, padding_mode=padding_mode),
            nn.BatchNorm2d(mid_channels),
            nn.ELU(inplace=True),
            nn.Conv2d(mid_channels, out_channels, kernel_size=3, padding=1, padding_mode=padding_mode),
            nn.BatchNorm2d(out_channels),
            nn.ELU(inplace=True)
        )

    def forward(self, x):
        return self.double_conv(x)


class Down(nn.Module):
    """Downscaling with maxpool then double conv"""

    def __init__(self, in_channels, out_channels, padding_mode='zeros'):
        super(Down, self).__init__()
        self.maxpool_conv = nn.Sequential(
            nn.MaxPool2d(2),
            DoubleConv(in_channels, out_channels)
        )

    def forward(self, x):
        return self.maxpool_conv(x)


class Up(nn.Module):
    """Upscaling then double conv"""

    def __init__(self, in_channels, out_channels, bilinear=True):
        super(Up, self).__init__()

        # if bilinear, use the normal convolutions to reduce the number of channels
        if bilinear:
            self.up = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)
            self.conv = DoubleConv(in_channels, out_channels, in_channels // 2)
        else:
            self.up = nn.ConvTranspose2d(in_channels , in_channels // 2, kernel_size=2, stride=2)
            self.conv = DoubleConv(in_channels, out_channels)


    def forward(self, x1, x2):
        x1 = self.up(x1)
        # input is CHW
        diffY = x2.size()[2] - x1.size()[2]
        diffX = x2.size()[3] - x1.size()[3]

        x1 = F.pad(x1, [diffX // 2, diffX - diffX // 2,
                        diffY // 2, diffY - diffY // 2], mode='replicate')
        # if you have padding issues, see
        # https://github.com/HaiyongJiang/U-Net-Pytorch-Unstructured-Buggy/commit/0e854509c2cea854e247a9c615f175f76fbb2e3a
        # https://github.com/xiaopeng-liao/Pytorch-UNet/commit/8ebac70e633bac59fc22bb5195e513d5832fb3bd
        x = torch.cat([x2, x1], dim=1)
        return self.conv(x)


class OutConv(nn.Module):
    def __init__(self, in_channels, out_channels, padding_mode='zeros'):
        super(OutConv, self).__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size=1, padding_mode=padding_mode)

    def forward(self, x):
        return self.conv(x)

class UNet(nn.Module):
    def __init__(self, n_channels, n_classes, bilinear=True, padding_mode='zeros'):
        super(UNet, self).__init__()
        self.n_channels = n_channels
        self.n_classes = n_classes
        self.bilinear = bilinear

        self.inc = DoubleConv(n_channels, 64, padding_mode=padding_mode)
        self.down1 = Down(64, 128, padding_mode=padding_mode)
        self.down2 = Down(128, 256, padding_mode=padding_mode)
        self.down3 = Down(256, 512, padding_mode=padding_mode)
        factor = 2 if bilinear else 1
        self.down4 = Down(512, 1024 // factor, padding_mode=padding_mode)
        self.up1 = Up(1024, 512 // factor, bilinear)
        self.up2 = Up(512, 256 // factor, bilinear)
        self.up3 = Up(256, 128 // factor, bilinear)
        self.up4 = Up(128, 64, bilinear)
        self.outc = OutConv(64, n_classes, padding_mode=padding_mode)

    def forward(self, x):
        x1 = self.inc(x)
        x2 = self.down1(x1)
        x3 = self.down2(x2)
        x4 = self.down3(x3)
        x5 = self.down4(x4)
        x = self.up1(x5, x4)
        x = self.up2(x, x3)
        x = self.up3(x, x2)
        x = self.up4(x, x1)
        logits = self.outc(x)
        return logits

if __name__ == "__main__":
    device = torch.device('cuda:0')

    dummy_in = torch.randn([8,1,512,512]).to(device)

    model = UNet(1,1).to(device)

    model.train()

    dummpy_out = model(dummy_in)

    print(dummpy_out.shape)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File  : sim.py
# Author: Alvin(Xinyao) Sun <xinyao1@ualberta.ca>
# Date  : 25.02.2021

import numpy as np
from .utils import wrap


def gen_sim_3d(mr,
               he,
               stack_length,
               bperp_scale=2000.,
               dday_stepmax=4,
               dday_scale=11,
               conv1_scale=-0.0000573803,
               conv1_shift=-0.0110171730107,
               conv2_scale=-0.00073405573,
               conv2_shift=-0.00086772422789899997):
    """gen_sim_3d.
    Generate simulated unwraped recon phase with given mr and he
    Args:
        mr: motion rate with shape [H, W]
        he: height error with shape [H, W]
        stack_length: length of stack size
    """
    ddays = np.random.randint(
        1, dday_stepmax, stack_length) * dday_scale  # shape: [stack_length]
    bperps = (np.random.rand(stack_length) -
              0.5) * bperp_scale  # shape: [stack_length]
    conv1 = np.random.rand() * (conv1_scale) + conv1_shift
    conv2 = np.random.rand() * (conv2_scale) + conv2_shift
    unwrap_recon_phase = conv1 * ddays * (np.expand_dims(
        mr, -1)) + conv2 * bperps * (np.expand_dims(he, -1)
                                    )  # shape: [H, W, stack_length]
    return unwrap_recon_phase, ddays, bperps, conv1, conv2
